/*<html><pre>*/
#include <../source/TAL/objbase.h>
#include <../source/PAL/dataman.h>
#include <../source/PAL/oshead.h>

#ifdef		__WIN32__

#define		delay_execution(milliseconds)		Sleep(milliseconds)
#define		get_tick_count						GetTickCount

#elif		defined(__UNIX__)	/* !__WIN32__ */

#ifdef		EMULATE_USLEEP
/* nap()'ing support routine for 4.2BSD system call emulations
 * Last edit:   29-Oct-1984    D A Gwyn
 * Returns 0 if ok, else -1 */
int	delay_execution(long milliseconds)
{
	static struct /* `timeval' */
	{
    	long tv_sec;      /* seconds */
        long tv_usec;     /* microsecs */
	}   delay;   /* _select() timeout */
	delay.tv_sec = microseconds / 1000000L;
    delay.tv_usec = microseconds % 1000000L;
	return select(0,(long *)0,(long *)0,(long *)0,&delay);
}
#else		/* !EMULATE_USLEEP */

#define		delay_execution(milliseconds)	usleep(1000*milliseconds)

#endif		/* !EMULATE_USLEEP */

u32_t	V_LOCAL
get_tick_count(void)
{
	struct timezone tz={0,0};
	struct timeval tv;
	gettimeofday((struct timeval*)&tv,&tz);
	return tv.tv_sec * 1000 + tv.tv_usec / 1000;
}

#endif	/* __UNIX__ */

u32_t	V_GLOBAL
timeTickCount()
{
	return get_tick_count();
}

void	V_GLOBAL
timeDelay(u32_t Milliseconds)
{
	delay_execution(Milliseconds);
}

/*
 *	Sets timer to active state.
 */
void	V_GLOBAL
timerStart(timer_t *timer)
{
	if(!timer)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,"timerStart(NULL): invalid handle.\n");
#endif	/* CHECKED_BUILD */
		return;
	}
	timer->active=true;
	timer->time_start=get_tick_count();
}

/*
 *	Sets timer to passive state: subsequent calls of timerMilliseconds()
 *	return time at the moment of this call.
 */	
void	V_GLOBAL
timerStop(timer_t *timer)
{
	if(!timer)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,"timerStop(NULL): invalid handle.\n");
#endif	/* CHECKED_BUILD */
		return;
	}
	timer->active=false;
	timer->time_start=get_tick_count();
}

u32_t	V_GLOBAL
timerMilliseconds(timer_t *timer)
{
	if(!timer)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,"timerMilliseconds(NULL): invalid handle.\n");
#endif	/* CHECKED_BUILD */
		return 0;
	}
	if(timer->active)
		return get_tick_count() - timer->time_start;
	else
		return timer->time_start;
}

void	V_GLOBAL
timerWait(timer_t *timer,u32_t return_time)
{
	u32_t delay_time;
    /* Let timerMilliseconds() check handle ... */
	delay_time=return_time-timerMilliseconds(timer);
	delay_execution(delay_time);
}

/*</pre></html>*/
/*EOF*/
