/*<html><pre>*/
#include <../source/TAL/objbase.h>
#include <../source/geom/stroke.h>

/*
#define DEBUG_STROKE
*/

/*
 *	Creates stroke object, copies coords. Assumes all args valid.
 */
V_LOCAL mesh_stroke_p
meshCreateStroke(vec3_t *coords,u32_t num_coords)
{
    mesh_stroke_p stroke;
	u32_t coord_len=num_coords*sizeof(vec3_t);
    stroke=(mesh_stroke_p)memAlloc(sizeof(u32_t)+coord_len);
    stroke->num_coords=num_coords;
    memCopy(stroke->coords,coords,coord_len);
    return stroke;
}

/*
 *	This object represent series of linestrips (polylines).
 */
typedef struct strokes_array_s
{
	libv_object_t base;
	bstack_t	coords;	 	/* Coordinates for current linestrip */
    vp_stack_t	strips; 	/* Linestrips for current glyph */
    vp_stack_t	glyphs;		/* Complete glyphs */
}	strokes_array_t, *strokes_array_p;

#ifdef	CHECKED_BUILD
#define	SARRAY_MAGIC		0x53415252
V_LOCAL char	*sarray_object="strokes array";
#endif	/* CHECKED_BUILD */

V_LOCAL u32_t V_CALL
saDestroy(object_handle_t handle)
{
    void *s, *g;
	strokes_array_t *sa=(strokes_array_t*)handle;
	if(--sa->base.ref_count)
		return sa->base.ref_count;
	/* Clear coords */
    stackClear(&sa->coords);
    /* Delete all strokes */
    while(vpsPop(&sa->strips,&s))
    	memFree(s);
	vpsClear(&sa->strips);
    /* Delete all glyphs */
    while(vpsPop(&sa->glyphs,&g))
    {
	    while(vpsPop((vp_stack_t *)g,&s))
    		memFree(s);
		vpsClear((vp_stack_t *)g);
        memFree(g);
	}
	vpsClear(&sa->glyphs);
    memFreeObject(handle);
	return 0;
}

object_handle_t	V_GLOBAL
meshCreateStrokes(void)
{
	strokes_array_t *sa;
    ALLOC_OBJECT(sa,strokes_array_t,0,SARRAY_MAGIC);
    sa->base.destroy=saDestroy;
    stackInit(&sa->coords,sizeof(vec3_t));
    vpsInit(&sa->strips);
    vpsInit(&sa->glyphs);
    return (object_handle_t)sa;
}

void	V_GLOBAL
meshDeleteStrokes(object_handle_t ms)
{
	libvCloseHandle(ms,"meshDeleteStrokes",sarray_object,SARRAY_MAGIC);
}

void	V_GLOBAL
meshStrokeCoord(object_handle_t ms,vec_t x,vec_t y,vec_t z)
{
	vec3_t v;
    vec3Set(v,x,y,z);
    meshStrokeCoordv(ms,v);
}

void	V_GLOBAL
meshStrokeCoordv(object_handle_t ms,vec3_t xyz)
{
	strokes_array_t *sa=(strokes_array_t*)
    	libvIsHandle(ms,"meshStrokeCoordv",sarray_object,SARRAY_MAGIC);
	if(!sa)
    	return;
	stackPush(&sa->coords,xyz);
}

void	V_GLOBAL
meshCancelStroke(object_handle_t ms)
{
	strokes_array_t *sa=(strokes_array_t*)
    	libvIsHandle(ms,"meshCancelStroke",sarray_object,SARRAY_MAGIC);
	if(!sa)
    	return;
	/* Clear coords */
    stackResize(&sa->coords,0);
}

void	V_GLOBAL
meshCancelGlyph(object_handle_t ms)
{
    void *s;
	strokes_array_t *sa=(strokes_array_t*)
    	libvIsHandle(ms,"meshCancelStroke",sarray_object,SARRAY_MAGIC);
	if(!sa)
    	return;
    /* Delete all strokes for new glyph */
    while(vpsPop(&sa->strips,&s))
    	memFree(s);
	/* Clear coords */
    stackResize(&sa->coords,0);
}

/*
 *	Stores recorded stroke in array
 */
void	V_GLOBAL
meshEndStroke(object_handle_t ms)
{
	u32_t count;
    vec3_t *coords;
    mesh_stroke_t *stroke;
    strokes_array_t *sa=(strokes_array_t *)
    	libvIsHandle(ms,"meshEndStroke",sarray_object,SARRAY_MAGIC);
	if(!sa)
    	return;
    count=stackCount(&sa->coords);
    if(count>1)
    {
    	coords=(vec3_t*)stackGetStorage(&sa->coords);
        stroke=meshCreateStroke(coords,count);
        if(stroke)
			vpsPush(&sa->strips,stroke);
#ifdef	CHECKED_BUILD
		else
        	conlPrintf(LOG_ERROR,"meshEndStroke(): "
            	"meshCreateStroke(%X,%u) failed.\n",
                coords,count);
#endif	/* CHECKED_BUILD */
	}
#if	0
	/* BGI fonts contains MoveTo just before end. */
    else
    {
    	conlPuts(LOG_WARNING,"meshEndStroke(): stroke must contain "
        	"at least 2 coords.\n");
    }
#endif	/* 0 */
	stackResize(&sa->coords,0);
    return;
}

/*
 *	Stores recorded array of strokes in array of glyphs
 */
void	V_GLOBAL
meshEndGlyph(object_handle_t ms)
{
	u32_t count;
    vp_stack_t *glyph;
    strokes_array_t *sa=(strokes_array_t *)
    	libvIsHandle(ms,"meshEndGlyph",sarray_object,SARRAY_MAGIC);
	if(!sa)
    	return;
    count=vpsCount(&sa->strips);
    /* Duplicate stack */
    glyph=(vp_stack_t*)memAlloc(sizeof(vp_stack_t));
    vpsInit(glyph);
    /* Note: it's Ok to create glyph without any strokes. */
    if(count)
    {
	    vpsConcat(glyph,&sa->strips);
        vpsAdjust(glyph);
	}
	vpsPush(&sa->glyphs,glyph);
	stackResize(&sa->coords,0);
	vpsResize(&sa->strips,0);
    return;
}

u32_t	V_GLOBAL
meshGlyphCount(object_handle_t ms)
{
    strokes_array_t *sa;
    sa=(strokes_array_t *)
    	libvIsHandle(ms,"meshGlyphCount",sarray_object,SARRAY_MAGIC);
	if(!sa)
    	return 0;
	return vpsCount(&sa->glyphs);
}

u32_t	V_GLOBAL
meshStrokesCount(object_handle_t ms,u32_t glyph_num)
{
    strokes_array_t *sa;
    sa=(strokes_array_t *)
    	libvIsHandle(ms,"meshStrokesCount",sarray_object,SARRAY_MAGIC);
	if(!sa)
    	return 0;
	if(glyph_num>=vpsCount(&sa->glyphs))
    {
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,"meshStrokesCount(%X,%u): no such glyph.\n",
        	ms,glyph_num);
#endif	/* CHECKED_BUILD */
    	return 0;
    }
	return vpsCount((vp_stack_t*)vpsGet(&sa->glyphs,glyph_num));
}

mesh_stroke_p	V_GLOBAL
meshGlyphStroke(object_handle_t ms,u32_t glyph_num,u32_t stroke_num)
{
	vp_stack_t *glyph;
    strokes_array_t *sa;
    sa=(strokes_array_t *)
    	libvIsHandle(ms,"meshGlyphStroke",sarray_object,SARRAY_MAGIC);
	if(!sa)
    	return NULL;
	if(glyph_num>=vpsCount(&sa->glyphs))
    {
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,"meshGlyphStroke(%X,%u,%u): no such glyph.\n",
        	ms,glyph_num,stroke_num);
#endif	/* CHECKED_BUILD */
    	return NULL;
    }
    glyph=(vp_stack_t*)vpsGet(&sa->glyphs,glyph_num);
	if(stroke_num>=vpsCount(glyph))
    {
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,"meshGlyphStroke(%X,%u,%u): no such stroke.\n",
        	ms,glyph_num,stroke_num);
#endif	/* CHECKED_BUILD */
    	return NULL;
    }
	return (mesh_stroke_p)vpsGet(glyph,stroke_num);
}

/*
**	Debugging stuff
*/

#ifdef	DEBUG_STROKE

void V_CALL
meshDumpStrokes(object_handle_t ms,FILE *out)
{
    strokes_array_t *sa;
    vp_stack_t *glyph;
    mesh_stroke_t *stroke;
	u32_t i, j, k, num_strokes, num_glyphs;
    sa=(strokes_array_t *)libvIsHandle(ms,"meshDumpStrokes",sarray_object,
    	SARRAY_MAGIC);
    if(!sa)
    	return;
	if(!out)
        out=stderr;
    if(stackCount(&sa->coords))
    	fprintf(out,
        	"Uncomplete stroke: %u coords.\n",stackCount(&sa->coords));
    if(vpsCount(&sa->strips))
    	fprintf(out,
        	"Uncomplete glyph: %u strokes.\n",vpsCount(&sa->strips));
    num_glyphs=vpsCount(&sa->glyphs);
	fprintf(out,"Glyphs complete: %u.\n",num_glyphs);
    /* Dump all strokes in complete glyphs */
    for(i=0;i<num_glyphs;i++)
    {
    	glyph=(vp_stack_t*)vpsGet(&sa->glyphs,i);
        num_strokes=vpsCount(glyph);
        fprintf(out,"Glyph %u: %u strokes.\n",i,num_strokes);
		for(j=0;j<num_strokes;j++)
	    {
        	stroke=(mesh_stroke_t*)vpsGet(glyph,j);
        	fprintf(out,"Stroke %u: %d coords: ",j,stroke->num_coords);
            for(k=0;k<stroke->num_coords;k++)
            	fprintf(out,"(%g,%g,%g); ",stroke->coords[k][0],
                	stroke->coords[k][1],stroke->coords[k][2]);
			fprintf(out,"\n");
		}
    }
}

void V_CALL
meshDumpStrokes2(object_handle_t ms,FILE *out)
{
    mesh_stroke_t *stroke;
	u32_t i, j, k, num_strokes, num_glyphs;
	if(!out)
        out=stderr;
    num_glyphs=meshGlyphCount(ms);
	fprintf(out,"Glyphs complete: %u.\n",num_glyphs);
    /* Dump all strokes in complete glyphs */
    for(i=0;i<num_glyphs;i++)
    {
    	num_strokes=meshStrokesCount(ms,i);
        fprintf(out,"Glyph %u: %u strokes.\n",i,num_strokes);
		for(j=0;j<num_strokes;j++)
	    {
        	stroke=meshGlyphStroke(ms,i,j);
        	if(!stroke)
        		continue;
        	fprintf(out,"Stroke %u: %d coords: ",j,stroke->num_coords);
            for(k=0;k<stroke->num_coords;k++)
            	fprintf(out,"(%g,%g,%g); ",stroke->coords[k][0],
                	stroke->coords[k][1],stroke->coords[k][2]);
			fprintf(out,"\n");
		}
    }
}

int main()
{
    object_handle_t ms;
    logStartConsoleLog();
    logUpperLevel(LOG_TRACE);
    ms=meshCreateStrokes();
    meshStrokeCoord(ms,1,2,0);
    meshStrokeCoord(ms,3,4,0);
    meshEndStroke(ms);
    meshStrokeCoord(ms,1,2,1);
    meshStrokeCoord(ms,3,4,1);
    meshEndStroke(ms);
    meshEndGlyph(ms);
    meshStrokeCoord(ms,0,0,1);
    meshStrokeCoord(ms,0,0,2);
    meshEndStroke(ms);
    meshEndGlyph(ms);
    meshStrokeCoord(ms,0,0,2);
    meshEndStroke(ms);
    meshStrokeCoord(ms,0,0,2);
    meshDumpStrokes(ms,stdout);
    meshDeleteStrokes(ms);
	return 0;
}

#endif	/* DEBUG_STROKE */

